@@ -25,12 +25,16 @@ njit = decorators.njit
 # Re export vectorize decorators
 from .npyufunc import vectorize, guvectorize
 
+# Re export from_dtype
+from .numpy_support import from_dtype
+
 # Re-export test entrypoint
 test = testing.test
 
 # Try initialize cuda
 from . import cuda
 
+
 __all__ = """
 jit
 autojit
@@ -40,31 +44,32 @@ guvectorize
 export
 exportmany
 cuda
+from_dtype
 """.split() + types.__all__ + special.__all__
 
 
 def _sentry_llvm_version():
     """
-    Make sure we meet min llvmpy version
+    Make sure we meet min llvmlite version
     """
     import warnings
-    import llvm
-    min_version = (0, 12, 6)
+    import llvmlite
+    min_version = (0, 1, 0)
 
     # Only look at the the major, minor and bugfix version numbers.
     # Ignore other stuffs
     regex = re.compile(r'(\d+)\.(\d+).(\d+)')
-    m = regex.match(llvm.__version__)
+    m = regex.match(llvmlite.__version__)
     if m:
         ver = tuple(map(int, m.groups()))
         if ver < min_version:
-            msg = ("Numba requires at least version %d.%d.%d of llvmpy.\n"
+            msg = ("Numba requires at least version %d.%d.%d of llvmlite.\n"
                    "Installed version is %s.\n"
-                   "Please update llvmpy." %
-                   (min_version + (llvm.__version__,)))
+                   "Please update llvmlite." %
+                   (min_version + (llvmlite.__version__,)))
             raise ImportError(msg)
     else:
         # Not matching?
-        warnings.warn("llvmpy version format not recognized!")
+        warnings.warn("llvmlite version format not recognized!")
 
 _sentry_llvm_version()
