@@ -2,6 +2,7 @@
 #define NUMBA_PY_MODULE_H_
 
 #include <Python.h>
+#include <structmember.h>
 
 #if PY_MAJOR_VERSION >= 3
   #define MOD_ERROR_VAL NULL
@@ -24,6 +25,7 @@
 
 #if (PY_MAJOR_VERSION >= 3)
     #define PyString_AsString PyUnicode_AsUTF8
+    #define PyString_Check PyUnicode_Check
     #define PyString_FromString PyUnicode_FromString
     #define PyString_InternFromString PyUnicode_InternFromString
     #define PyInt_Type PyLong_Type
