@@ -4,11 +4,10 @@ from contextlib import contextmanager
 from collections import namedtuple, defaultdict
 import warnings
 import inspect
-from llvmlite import binding as ll
 
 from numba import (bytecode, interpreter, typing, typeinfer, lowering,
                    objmode, irpasses, utils, config, type_annotations,
-                   types, ir, assume, looplifting, macro)
+                   types, ir, assume, looplifting, macro, types)
 from numba.targets import cpu
 
 
@@ -25,10 +24,13 @@ class Flags(utils.ConfigOptions):
         'enable_pyobject_looplift',
         # Force pyobject mode inside the whole function
         'force_pyobject',
+        # Release GIL inside the native function
+        'release_gil',
         'no_compile',
         'boundcheck',
+        'forceinline',
         'no_cpython_wrapper',
-        ])
+    ])
 
 
 DEFAULT_FLAGS = Flags()
@@ -45,15 +47,24 @@ CR_FIELDS = ["typing_context",
              "fndesc",
              "interpmode",
              "library",
-             "exception_map"]
+             "exception_map",
+             "environment"]
 
 
 CompileResult = namedtuple("CompileResult", CR_FIELDS)
 FunctionAttributes = namedtuple("FunctionAttributes",
-    ['name', 'filename', 'lineno'])
+                                ['name', 'filename', 'lineno'])
 DEFAULT_FUNCTION_ATTRIBUTES = FunctionAttributes('<anonymous>', '<unknown>', 0)
 
 
+_LowerResult = namedtuple("_LowerResult", [
+    "fndesc",
+    "exception_map",
+    "cfunc",
+    "env",
+])
+
+
 def get_function_attributes(func):
     '''
     Extract the function attributes from a Python function or object with
@@ -128,54 +139,65 @@ class _CompileStatus(object):
         return ', '.join(vals)
 
 
-class _StageResult(object):
-    def __init__(self, stage, ok, result=None, exception=None,
-                 early_escape=False):
-        self.stage = stage
-        self.ok = ok
-        self.early_escape = early_escape
-        self._result = result
-        self._exception = exception
-
-    @property
-    def result(self):
-        assert self.ok
-        return self._result
-
-    @property
-    def exception(self):
-        assert not self.ok
-        return self._exception
-
-
-class _EarlyEscape(BaseException):
+class _EarlyPipelineCompletion(Exception):
     def __init__(self, result):
         self.result = result
 
 
-class _Stage(object):
-    def __init__(self, func, msg):
-        self.func = func
-        self.msg = msg
+class _PipelineManager(object):
+    def __init__(self):
+        self.pipeline_order = []
+        self.pipeline_stages = {}
+        self._finalized = False
 
-    def __call__(self):
-        try:
-            res = self.func()
-        except _EarlyEscape as e:
-            return _StageResult(self, ok=True, result=e.result,
-                               early_escape=True)
-        except BaseException as e:
-            return _StageResult(self, ok=False, exception=e)
-        else:
-            return _StageResult(self, ok=True, result=res)
+    def create_pipeline(self, pipeline_name):
+        assert not self._finalized, "Pipelines can no longer be added"
+        self.pipeline_order.append(pipeline_name)
+        self.pipeline_stages[pipeline_name] = []
+        self.current = pipeline_name
 
+    def add_stage(self, stage_function, stage_description):
+        assert not self._finalized, "Stages can no longer be added."
+        current_pipeline_name = self.pipeline_order[-1]
+        func_desc_tuple = (stage_function, stage_description)
+        self.pipeline_stages[current_pipeline_name].append(func_desc_tuple)
 
-def _raise_error(desc, exc):
-    """Patches the error
-    """
-    newmsg = "{desc}\n{exc}".format(desc=desc, exc=exc)
-    exc.args = (newmsg,)
-    return exc
+    def finalize(self):
+        self._finalized = True
+
+    def _patch_error(self, desc, exc):
+        """
+        Patches the error to show the stage that it arose in.
+        """
+        newmsg = "{desc}\n{exc}".format(desc=desc, exc=exc)
+        exc.args = (newmsg,)
+        return exc
+
+    def run(self, status):
+        assert self._finalized, "PM must be finalized before run()"
+        res = None
+        for pipeline_name in self.pipeline_order:
+            is_final_pipeline = pipeline_name == self.pipeline_order[-1]
+            for stage, stage_name in self.pipeline_stages[pipeline_name]:
+                try:
+                    res = stage()
+                except _EarlyPipelineCompletion as e:
+                    return e.result
+                except BaseException as e:
+                    msg = "Failed at %s (%s)" % (pipeline_name, stage_name)
+                    patched_exception = self._patch_error(msg, e)
+                    # No more fallback pipelines?
+                    if is_final_pipeline:
+                        raise patched_exception
+                    # Go to next fallback pipeline
+                    else:
+                        status.fail_reason = patched_exception
+                        break
+            else:
+                return res
+
+        # TODO save all error information
+        raise CompilerError("All pipelines have failed")
 
 
 class CompilerError(Exception):
@@ -183,7 +205,8 @@ class CompilerError(Exception):
 
 
 class Pipeline(object):
-    """Stores and manages states for the compiler pipeline
+    """
+    Stores and manages states for the compiler pipeline
     """
     def __init__(self, typingctx, targetctx, library, args, return_type, flags,
                  locals):
@@ -205,7 +228,8 @@ class Pipeline(object):
 
     @contextmanager
     def fallback_context(self, msg):
-        """Wraps code that would signal a fallback to object mode
+        """
+        Wraps code that would signal a fallback to object mode
         """
         try:
             yield
@@ -225,7 +249,8 @@ class Pipeline(object):
 
     @contextmanager
     def giveup_context(self, msg):
-        """Wraps code that would signal a fallback to interpreter mode
+        """
+        Wraps code that would signal a fallback to interpreter mode
         """
         try:
             yield
@@ -257,14 +282,15 @@ class Pipeline(object):
         return bc
 
     def compile_extra(self, func):
-        res = _Stage(lambda: self.extract_bytecode(func),
-                    "extract bytecode")()
-        if res.ok:
-            return self.compile_bytecode(res.result, func_attr=self.func_attr)
-        elif self.status.can_giveup:
-            return self.stage_compile_interp_mode()
-        else:
-            raise res.exception
+        try:
+            bc = self.extract_bytecode(func)
+        except BaseException as e:
+            if self.status.can_giveup:
+                return self.stage_compile_interp_mode()
+            else:
+                raise e
+
+        return self.compile_bytecode(bc, func_attr=self.func_attr)
 
     def compile_bytecode(self, bc, lifted=(),
                          func_attr=DEFAULT_FUNCTION_ATTRIBUTES):
@@ -289,8 +315,14 @@ class Pipeline(object):
         """
         self.interp = translate_stage(self.bc)
         self.nargs = len(self.interp.argspec.args)
-        if len(self.args) > self.nargs:
-            raise TypeError("Too many argument types")
+        if not self.args and self.flags.force_pyobject:
+            # Allow an empty argument types specification when object mode
+            # is explicitly requested.
+            self.args = (types.pyobject,) * self.nargs
+        elif len(self.args) != self.nargs:
+            raise TypeError("Signature mismatch: %d argument types given, "
+                            "but function takes %d arguments"
+                            % (len(self.args), self.nargs))
 
     def frontend_looplift(self):
         """
@@ -334,7 +366,7 @@ class Pipeline(object):
             assert not self.lifted
             cres = self.frontend_looplift()
             if cres is not None:
-                raise _EarlyEscape(cres)
+                raise _EarlyPipelineCompletion(cres)
 
         # Fallback typing: everything is a python object
         self.typemap = defaultdict(lambda: types.pyobject)
@@ -377,7 +409,9 @@ class Pipeline(object):
             print('=' * 80)
 
     def backend_object_mode(self):
-        """Object mode compilation"""
+        """
+        Object mode compilation
+        """
         with self.giveup_context("Function %s failed at object mode lowering"
                                  % (self.func_attr.name,)):
             if len(self.args) != self.nargs:
@@ -410,20 +444,21 @@ class Pipeline(object):
         if self.library is None:
             codegen = self.targetctx.jit_codegen()
             self.library = codegen.create_library(self.bc.func_qualname)
-        fndesc, exception_map, func = lowerfn()
+        lowered = lowerfn()
         signature = typing.signature(self.return_type, *self.args)
         cr = compile_result(typing_context=self.typingctx,
                             target_context=self.targetctx,
-                            entry_point=func,
+                            entry_point=lowered.cfunc,
                             typing_error=self.status.fail_reason,
                             type_annotation=self.type_annotation,
                             library=self.library,
-                            exception_map=exception_map,
+                            exception_map=lowered.exception_map,
                             signature=signature,
                             objectmode=objectmode,
                             interpmode=False,
                             lifted=self.lifted,
-                            fndesc=fndesc,)
+                            fndesc=lowered.fndesc,
+                            environment=lowered.env,)
         return cr
 
     def stage_objectmode_backend(self):
@@ -436,12 +471,17 @@ class Pipeline(object):
         # Warn if compiled function in object mode and force_pyobject not set
         if not self.flags.force_pyobject:
             if len(self.lifted) > 0:
-                warn_msg = 'Function "%s" was compiled in object mode without forceobj=True, but has lifted loops.' % self.func_attr.name,
+                warn_msg = 'Function "%s" was compiled in object mode without forceobj=True, but has lifted loops.' % (self.func_attr.name,)
             else:
-                warn_msg = 'Function "%s" was compiled in object mode without forceobj=True.' % self.func_attr.name,
+                warn_msg = 'Function "%s" was compiled in object mode without forceobj=True.' % (self.func_attr.name,)
             warnings.warn_explicit(warn_msg, config.NumbaWarning,
                                    self.func_attr.filename,
                                    self.func_attr.lineno)
+            if self.flags.release_gil:
+                warn_msg = "Code running in object mode won't allow parallel execution despite nogil=True."
+                warnings.warn_explicit(warn_msg, config.NumbaWarning,
+                                       self.func_attr.filename,
+                                       self.func_attr.lineno)
         return res
 
     def stage_nopython_backend(self):
@@ -470,73 +510,28 @@ class Pipeline(object):
         return cr
 
     def _compile_bytecode(self):
-        pipelines = []
+        pm = _PipelineManager()
 
         if not self.flags.force_pyobject:
-            nopython_stages = [
-                _Stage(self.stage_analyze_bytecode, "analyzing bytecode"),
-                _Stage(self.stage_nopython_frontend, "nopython frontend"),
-                _Stage(self.stage_annotate_type, "annotate type"),
-                _Stage(self.stage_nopython_backend, "nopython mode backend"),
-            ]
-            pipelines.append(nopython_stages)
+            pm.create_pipeline("nopython")
+            pm.add_stage(self.stage_analyze_bytecode, "analyzing bytecode")
+            pm.add_stage(self.stage_nopython_frontend, "nopython frontend")
+            pm.add_stage(self.stage_annotate_type, "annotate type")
+            pm.add_stage(self.stage_nopython_backend, "nopython mode backend")
 
         if self.status.can_fallback or self.flags.force_pyobject:
-            object_stages = [
-                _Stage(self.stage_analyze_bytecode, "analyzing bytecode"),
-                _Stage(self.stage_objectmode_frontend, "object mode frontend"),
-                _Stage(self.stage_annotate_type, "annotate type"),
-                _Stage(self.stage_objectmode_backend, "object mode backend")
-            ]
-            pipelines.append(object_stages)
+            pm.create_pipeline("object")
+            pm.add_stage(self.stage_analyze_bytecode, "analyzing bytecode")
+            pm.add_stage(self.stage_objectmode_frontend, "object mode frontend")
+            pm.add_stage(self.stage_annotate_type, "annotate type")
+            pm.add_stage(self.stage_objectmode_backend, "object mode backend")
 
         if self.status.can_giveup:
-            interp_stages = [
-                _Stage(self.stage_compile_interp_mode,
-                      "compiling with interpreter mode"),
-            ]
-            pipelines.append(interp_stages)
+            pm.create_pipeline("interp")
+            pm.add_stage(self.stage_compile_interp_mode, "compiling with interpreter mode")
 
-        assert pipelines
-        return self._run_pipeline(pipelines)
-
-    def _run_pipeline(self, pipelines):
-        """
-        Args
-        -----
-        pipelines : sequence of sequence of Stage
-            Multiple pipeline of of Stages to execute.
-            The first pipeline is attempted first.
-            If it fails, the next pipeline is tried.
-            If all pipelines fail, an error is showed
-
-        Returns
-        -------
-        The result of the last Stage.
-        """
-        res = None
-        for pi, stages in enumerate(pipelines):
-            for stage in stages:
-                res = stage()
-                # Stage failed?
-                if not res.ok:
-                    # No more fallback pipelines?
-                    msg = "Failed at " + stage.msg
-                    if pi + 1 >= len(pipelines):
-                        raise _raise_error(msg, res.exception)
-                    # Go to next fallback pipeline
-                    else:
-                        self.status.fail_reason = _raise_error(msg,
-                                                               res.exception)
-                        break
-                # Stage OK and early escape
-                elif res.early_escape:
-                    return res.result
-            else:
-                return res.result
-
-        # TODO save all error information
-        raise CompilerError("All pipelines have failed")
+        pm.finalize()
+        return pm.run(self.status)
 
 
 def compile_extra(typingctx, targetctx, func, args, return_type, flags,
@@ -642,7 +637,7 @@ def translate_stage(bytecode):
 
     if config.DUMP_IR and expanded:
         print(("MACRO-EXPANDED IR DUMP: %s" % interp.bytecode.func_qualname)
-            .center(80, "-"))
+              .center(80, "-"))
         interp.dump()
 
     return interp
@@ -682,37 +677,44 @@ def native_lowering_stage(targetctx, library, interp, typemap, restype,
                           calltypes, flags):
     # Lowering
     fndesc = lowering.PythonFunctionDescriptor.from_specialized_function(
-        interp, typemap, restype, calltypes, mangler=targetctx.mangler)
+        interp, typemap, restype, calltypes, mangler=targetctx.mangler,
+        inline=flags.forceinline)
 
     lower = lowering.Lower(targetctx, library, fndesc, interp)
-    lower.lower(create_wrapper=not flags.no_cpython_wrapper)
+    lower.lower()
+    if not flags.no_cpython_wrapper:
+        lower.create_cpython_wrapper(flags.release_gil)
     env = lower.env
     exception_map = lower.exceptions
     del lower
 
     if flags.no_compile:
-        return fndesc, exception_map, None
+        return _LowerResult(fndesc, exception_map, cfunc=None, env=None)
     else:
         # Prepare for execution
         cfunc = targetctx.get_executable(library, fndesc, env)
-        targetctx.insert_user_function(cfunc, fndesc)
-        return fndesc, exception_map, cfunc
+        # Insert native function for use by other jitted-functions.
+        # We also register its library to allow for inlining.
+        targetctx.insert_user_function(cfunc, fndesc, [library])
+        return _LowerResult(fndesc, exception_map, cfunc=cfunc, env=None)
 
 
 def py_lowering_stage(targetctx, library, interp, flags):
     fndesc = lowering.PythonFunctionDescriptor.from_object_mode_function(interp)
     lower = objmode.PyLower(targetctx, library, fndesc, interp)
-    lower.lower(create_wrapper=not flags.no_cpython_wrapper)
+    lower.lower()
+    if not flags.no_cpython_wrapper:
+        lower.create_cpython_wrapper()
     env = lower.env
     exception_map = lower.exceptions
     del lower
 
     if flags.no_compile:
-        return fndesc, exception_map, None
+        return _LowerResult(fndesc, exception_map, cfunc=None, env=env)
     else:
         # Prepare for execution
         cfunc = targetctx.get_executable(library, fndesc, env)
-        return fndesc, exception_map, cfunc
+        return _LowerResult(fndesc, exception_map, cfunc, env)
 
 
 def ir_optimize_for_py_stage(interp):
