@@ -1,14 +1,20 @@
 from __future__ import print_function, division, absolute_import
+
 import struct
 import sys
 import os
 import re
 import warnings
 
+import llvmlite.binding as ll
+
+
 IS_WIN32 = sys.platform.startswith('win32')
 MACHINE_BITS = tuple.__itemsize__ * 8
 IS_32BITS = MACHINE_BITS == 32
 
+_cpu_name = ll.get_host_cpu_name()
+
 
 class NumbaWarning(Warning):
     pass
@@ -59,7 +65,7 @@ DUMP_IR = _readenv("NUMBA_DUMP_IR", int, DEBUG_FRONTEND)
 DUMP_LLVM = _readenv("NUMBA_DUMP_LLVM", int, DEBUG)
 
 # Force dump of Function optimized LLVM IR
-NUMBA_DUMP_FUNC_OPT = _readenv("NUMBA_DUMP_FUNC_OPT", int, DEBUG)
+DUMP_FUNC_OPT = _readenv("NUMBA_DUMP_FUNC_OPT", int, DEBUG)
 
 # Force dump of Optimized LLVM IR
 DUMP_OPTIMIZED = _readenv("NUMBA_DUMP_OPTIMIZED", int, DEBUG)
@@ -103,5 +109,6 @@ def _force_cc(text):
 FORCE_CUDA_CC = _readenv("NUMBA_FORCE_CUDA_CC", _force_cc, None)
 
 # x86-64 specific
-# Enable AVX on supported platforms.
-ENABLE_AVX = _readenv("NUMBA_ENABLE_AVX", int, 0)
+# Enable AVX on supported platforms where it won't degrade performance.
+ENABLE_AVX = _readenv("NUMBA_ENABLE_AVX", int,
+                      _cpu_name not in ('corei7-avx', 'core-avx-i'))
