@@ -17,6 +17,7 @@ except NameError:
 
 require_context = devices.require_context
 current_context = devices.get_context
+gpus = devices.gpus
 
 
 @require_context
@@ -186,7 +187,7 @@ def mapped(*arylist, **kws):
     pmlist = []
     stream = kws.get('stream', 0)
     for ary in arylist:
-        pm = current_context.mempin(ary, driver.host_pointer(ary),
+        pm = current_context().mempin(ary, driver.host_pointer(ary),
                                     driver.host_memory_size(ary),
                                     mapped=True)
         pmlist.append(pm)
@@ -212,7 +213,7 @@ def event(timing=True):
 def select_device(device_id):
     """Creates a new CUDA context with the selected device.
     The context is associated with the current thread.
-    NumbaPro currently allows only one context per thread.
+    Numba currently allows only one context per thread.
 
     Returns a device instance
 
@@ -229,7 +230,6 @@ def get_current_device():
 
 def list_devices():
     "List all CUDA devices"
-    devices.init_gpus()
     return devices.gpus
 
 
