@@ -77,7 +77,8 @@ def make_arrayiter_cls(iterator_type):
     """
 
     class ArrayIteratorStruct(cgutils.Structure):
-        _fields = [('index', types.CPointer(types.intp)),
+        # We use an unsigned index to avoid the cost of negative index tests.
+        _fields = [('index', types.CPointer(types.uintp)),
                    ('array', iterator_type.array_type)]
 
     return ArrayIteratorStruct
@@ -99,9 +100,9 @@ def getiter_array(context, builder, sig, args):
     return iterobj._getvalue()
 
 
-def _getitem_array1d(context, builder, arrayty, array, idx):
+def _getitem_array1d(context, builder, arrayty, array, idx, wraparound):
     ptr = cgutils.get_item_pointer(builder, arrayty, array, [idx],
-                                   wraparound=True)
+                                   wraparound=wraparound)
     return context.unpack_value(builder, arrayty.dtype, ptr)
 
 @builtin
@@ -126,15 +127,16 @@ def iternext_array(context, builder, sig, args, result):
     result.set_valid(is_valid)
 
     with cgutils.ifthen(builder, is_valid):
-        value = _getitem_array1d(context, builder, arrayty, ary, index)
+        value = _getitem_array1d(context, builder, arrayty, ary, index,
+                                 wraparound=False)
         result.yield_(value)
         nindex = builder.add(index, context.get_constant(types.intp, 1))
         builder.store(nindex, iterobj.index)
 
 @builtin
-@implement('getitem', types.Kind(types.Array), types.intp)
+@implement('getitem', types.Kind(types.Array), types.Kind(types.Integer))
 def getitem_array1d_intp(context, builder, sig, args):
-    aryty, _ = sig.args
+    aryty, idxty = sig.args
     if aryty.ndim != 1:
         # TODO
         raise NotImplementedError("1D indexing into %dD array" % aryty.ndim)
@@ -143,7 +145,8 @@ def getitem_array1d_intp(context, builder, sig, args):
 
     arystty = make_array(aryty)
     ary = arystty(context, builder, ary)
-    return _getitem_array1d(context, builder, aryty, ary, idx)
+    return _getitem_array1d(context, builder, aryty, ary, idx,
+                            wraparound=idxty.signed)
 
 @builtin
 @implement('getitem', types.Kind(types.Array), types.slice3_type)
@@ -218,12 +221,12 @@ def getitem_array_unituple(context, builder, sig, args):
         return retary._getvalue()
     else:
         # Indexing
-        assert idxty.dtype == types.intp
+        assert isinstance(idxty.dtype, types.Integer)
         indices = cgutils.unpack_tuple(builder, idx, count=len(idxty))
         indices = [context.cast(builder, i, t, types.intp)
                    for t, i in zip(idxty, indices)]
         ptr = cgutils.get_item_pointer(builder, aryty, ary, indices,
-                                       wraparound=True)
+                                       wraparound=idxty.dtype.signed)
 
         return context.unpack_value(builder, aryty.dtype, ptr)
 
@@ -281,17 +284,17 @@ def getitem_array_tuple(context, builder, sig, args):
 
 
 @builtin
-@implement('setitem', types.Kind(types.Array), types.intp,
+@implement('setitem', types.Kind(types.Array), types.Kind(types.Integer),
            types.Any)
 def setitem_array1d(context, builder, sig, args):
-    aryty, _, valty = sig.args
+    aryty, idxty, valty = sig.args
     ary, idx, val = args
 
     arystty = make_array(aryty)
     ary = arystty(context, builder, ary)
 
     ptr = cgutils.get_item_pointer(builder, aryty, ary, [idx],
-                                   wraparound=True)
+                                   wraparound=idxty.signed)
 
     val = context.cast(builder, val, valty, aryty.dtype)
 
@@ -313,7 +316,7 @@ def setitem_array_unituple(context, builder, sig, args):
     indices = [context.cast(builder, i, t, types.intp)
                for t, i in zip(idxty, indices)]
     ptr = cgutils.get_item_pointer(builder, aryty, ary, indices,
-                                   wraparound=True)
+                                   wraparound=idxty.dtype.signed)
     context.pack_value(builder, aryty.dtype, val, ptr)
 
 
@@ -424,79 +427,152 @@ def array_len(context, builder, sig, args):
 
 
 @builtin
+@implement(numpy.sum, types.Kind(types.Array))
 @implement("array.sum", types.Kind(types.Array))
 def array_sum(context, builder, sig, args):
-    from numba.intrinsics import array_ravel
-
     [arrty] = sig.args
 
-    def impl_any_layout(arr):
+    def array_sum_impl(arr):
         c = 0
         for v in arr.flat:
             c += v
         return c
 
-    def impl_contigous_layout(arr):
-        c = 0
-        for v in array_ravel(arr):
-            c += v
-        return c
-
-    if arrty.layout in 'CF':
-        # Optimize for contiguous case because so that LLVM can perform
-        # vectorization on the reduction loop
-        return context.compile_internal(builder, impl_contigous_layout, sig,
-                                        args, locals=dict(c=arrty.dtype))
-    else:
-        return context.compile_internal(builder, impl_any_layout, sig, args,
-                                        locals=dict(c=arrty.dtype))
-
-
-@builtin
-@implement(numpy.sum, types.Kind(types.Array))
-def numpy_sum(context, builder, sig, args):
-    def impl(arr):
-        return arr.sum()
-
-    return context.compile_internal(builder, impl, sig, args)
-
-
-@builtin
-@implement("array.prod", types.Kind(types.Array))
-def array_prod(context, builder, sig, args):
-    from numba.intrinsics import array_ravel
-
-    [arrty] = sig.args
-
-    def impl_any_layout(arr):
-        c = 1
-        for v in arr.flat:
-            c *= v
-        return c
-
-    def impl_contigous_layout(arr):
-        c = 1
-        for v in array_ravel(arr):
-            c *= v
-        return c
-
-    if arrty.layout in 'CF':
-        # Optimize for contiguous case because so that LLVM can perform
-        # vectorization on the reduction loop
-        return context.compile_internal(builder, impl_contigous_layout, sig,
-                                        args, locals=dict(c=arrty.dtype))
-    else:
-        return context.compile_internal(builder, impl_any_layout, sig, args,
-                                        locals=dict(c=arrty.dtype))
+    return context.compile_internal(builder, array_sum_impl, sig, args, 
+                                    locals=dict(c=sig.return_type))
 
 
 @builtin
 @implement(numpy.prod, types.Kind(types.Array))
-def numpy_prod(context, builder, sig, args):
-    def impl(arr):
-        return arr.prod()
+@implement("array.prod", types.Kind(types.Array))
+def array_prod(context, builder, sig, args):
+    [arrty] = sig.args
+
+    def array_prod_impl(arr):
+        c = 1
+        for v in arr.flat:
+            c *= v
+        return c
+
+    return context.compile_internal(builder, array_prod_impl, sig, args,
+                                    locals=dict(c=sig.return_type))
+
+
+@builtin
+@implement(numpy.mean, types.Kind(types.Array))
+@implement("array.mean", types.Kind(types.Array))
+def array_mean(context, builder, sig, args):
+    [arrty] = sig.args
+
+    def array_mean_impl(arr):
+        # Can't use the naive `arr.sum() / arr.size`, as it would return
+        # a wrong result on integer sum overflow.
+        c = 0
+        for v in arr.flat:
+            c += v
+        return c / arr.size
+
+    return context.compile_internal(builder, array_mean_impl, sig, args,
+                                    locals=dict(c=sig.return_type))
+
+
+@builtin
+@implement(numpy.var, types.Kind(types.Array))
+@implement("array.var", types.Kind(types.Array))
+def array_var(context, builder, sig, args):
+    def array_var_impl(arry):
+        # Compute the mean
+        m = arry.mean()
+
+        # Compute the sum of square diffs
+        ssd = 0
+        for v in arry.flat:
+            ssd += (v - m) ** 2
+        return ssd / arry.size
+
+    return context.compile_internal(builder, array_var_impl, sig, args)
+
+
+@builtin
+@implement(numpy.std, types.Kind(types.Array))
+@implement("array.std", types.Kind(types.Array))
+def array_std(context, builder, sig, args):
+    def array_std_impl(arry):
+        return arry.var() ** 0.5
+    return context.compile_internal(builder, array_std_impl, sig, args)
+
+
+@builtin
+@implement(numpy.min, types.Kind(types.Array))
+@implement("array.min", types.Kind(types.Array))
+def array_min(context, builder, sig, args):
+    def array_min_impl(arry):
+        for v in arry.flat:
+            min_value = v
+            break
+
+        for v in arry.flat:
+            if v < min_value:
+                min_value = v
+        return min_value
+    return context.compile_internal(builder, array_min_impl, sig, args)
+
+
+@builtin
+@implement(numpy.max, types.Kind(types.Array))
+@implement("array.max", types.Kind(types.Array))
+def array_max(context, builder, sig, args):
+    def array_max_impl(arry):
+        for v in arry.flat:
+            max_value = v
+            break
+        
+        for v in arry.flat:
+            if v > max_value:
+                max_value = v
+        return max_value
+    return context.compile_internal(builder, array_max_impl, sig, args)
+
+
+@builtin
+@implement(numpy.argmin, types.Kind(types.Array))
+@implement("array.argmin", types.Kind(types.Array))
+def array_argmin(context, builder, sig, args):
+    def array_argmin_impl(arry):
+        for v in arry.flat:
+            min_value = v
+            min_idx = 0
+            break
+
+        idx = 0
+        for v in arry.flat:
+            if v < min_value:
+                min_value = v
+                min_idx = idx
+            idx += 1
+        return min_idx
+    return context.compile_internal(builder, array_argmin_impl, sig, args)
+
+
+@builtin
+@implement(numpy.argmax, types.Kind(types.Array))
+@implement("array.argmax", types.Kind(types.Array))
+def array_argmax(context, builder, sig, args):
+    def array_argmax_impl(arry):
+        for v in arry.flat:
+            max_value = v
+            max_idx = 0
+            break
+
+        idx = 0
+        for v in arry.flat:
+            if v > max_value:
+                max_value = v
+                max_idx = idx
+            idx += 1
+        return max_idx
+    return context.compile_internal(builder, array_argmax_impl, sig, args)
 
-    return context.compile_internal(builder, impl, sig, args)
 
 #-------------------------------------------------------------------------------
 
@@ -574,6 +650,116 @@ def make_array_flat_cls(flatiterty):
     Return the Structure representation of the given *flatiterty* (an
     instance of types.NumpyFlatType).
     """
+    return _make_flattening_iter_cls(flatiterty, 'flat')
+
+
+@struct_factory(types.NumpyNdEnumerateType)
+def make_array_ndenumerate_cls(nditerty):
+    """
+    Return the Structure representation of the given *nditerty* (an
+    instance of types.NumpyNdEnumerateType).
+    """
+    return _make_flattening_iter_cls(nditerty, 'ndenumerate')
+
+
+def _increment_indices(context, builder, ndim, shape, indices, end_flag=None):
+    zero = context.get_constant(types.intp, 0)
+    one = context.get_constant(types.intp, 1)
+
+    bbend = cgutils.append_basic_block(builder, 'end_increment')
+
+    if end_flag is not None:
+        builder.store(cgutils.false_byte, end_flag)
+
+    for dim in reversed(range(ndim)):
+        idxptr = cgutils.gep(builder, indices, dim)
+        idx = builder.add(builder.load(idxptr), one)
+
+        count = shape[dim]
+        in_bounds = builder.icmp(lc.ICMP_SLT, idx, count)
+        with cgutils.if_likely(builder, in_bounds):
+            builder.store(idx, idxptr)
+            builder.branch(bbend)
+        builder.store(zero, idxptr)
+
+    if end_flag is not None:
+        builder.store(cgutils.true_byte, end_flag)
+    builder.branch(bbend)
+
+    builder.position_at_end(bbend)
+
+def _increment_indices_array(context, builder, arrty, arr, indices, end_flag=None):
+    shape = cgutils.unpack_tuple(builder, arr.shape, arrty.ndim)
+    _increment_indices(context, builder, arrty.ndim, shape, indices, end_flag)
+
+
+@struct_factory(types.NumpyNdIndexType)
+def make_ndindex_cls(nditerty):
+    """
+    Return the Structure representation of the given *nditerty* (an
+    instance of types.NumpyNdIndexType).
+    """
+    ndim = nditerty.ndim
+
+    class NdIndexIter(cgutils.Structure):
+        """
+        .ndindex() implementation.
+        """
+        _fields = [('shape', types.UniTuple(types.intp, ndim)),
+                   ('indices', types.CPointer(types.intp)),
+                   ('exhausted', types.CPointer(types.boolean)),
+                   ]
+
+        def init_specific(self, context, builder, shapes):
+            zero = context.get_constant(types.intp, 0)
+            indices = cgutils.alloca_once(builder, zero.type,
+                                          size=context.get_constant(types.intp,
+                                                                    ndim))
+            exhausted = cgutils.alloca_once_value(builder, cgutils.false_byte)
+
+            for dim in range(ndim):
+                idxptr = cgutils.gep(builder, indices, dim)
+                builder.store(zero, idxptr)
+                # 0-sized dimensions really indicate an empty array,
+                # but we have to catch that condition early to avoid
+                # a bug inside the iteration logic.
+                dim_size = shapes[dim]
+                dim_is_empty = builder.icmp(lc.ICMP_EQ, dim_size, zero)
+                with cgutils.if_unlikely(builder, dim_is_empty):
+                    builder.store(cgutils.true_byte, exhausted)
+
+            self.indices = indices
+            self.exhausted = exhausted
+            self.shape = cgutils.pack_array(builder, shapes)
+
+        def iternext_specific(self, context, builder, result):
+            zero = context.get_constant(types.intp, 0)
+            one = context.get_constant(types.intp, 1)
+
+            bbend = cgutils.append_basic_block(builder, 'end')
+
+            exhausted = cgutils.as_bool_bit(builder, builder.load(self.exhausted))
+            with cgutils.if_unlikely(builder, exhausted):
+                result.set_valid(False)
+                builder.branch(bbend)
+
+            indices = [builder.load(cgutils.gep(builder, self.indices, dim))
+                       for dim in range(ndim)]
+            result.yield_(cgutils.pack_array(builder, indices))
+            result.set_valid(True)
+
+            shape = cgutils.unpack_tuple(builder, self.shape, ndim)
+            _increment_indices(context, builder, ndim, shape,
+                               self.indices, self.exhausted)
+
+            builder.branch(bbend)
+            builder.position_at_end(bbend)
+
+    return NdIndexIter
+
+
+def _make_flattening_iter_cls(flatiterty, kind):
+    assert kind in ('flat', 'ndenumerate')
 
     array_type = flatiterty.array_type
     dtype = array_type.dtype
@@ -581,12 +767,13 @@ def make_array_flat_cls(flatiterty):
     if array_type.layout == 'C':
         class CContiguousFlatIter(cgutils.Structure):
             """
-            .flat() implementation for C-contiguous arrays.
+            .flat() / .ndenumerate() implementation for C-contiguous arrays.
             """
             _fields = [('array', types.CPointer(array_type)),
                        ('stride', types.intp),
                        ('pointer', types.CPointer(types.CPointer(dtype))),
                        ('index', types.CPointer(types.intp)),
+                       ('indices', types.CPointer(types.intp)),
                        ]
 
             def init_specific(self, context, builder, arrty, arr):
@@ -598,7 +785,23 @@ def make_array_flat_cls(flatiterty):
                 # http://docs.scipy.org/doc/numpy-dev/release.html#npy-relaxed-strides-checking
                 self.stride = arr.itemsize
 
+                if kind == 'ndenumerate':
+                    # Zero-initialize the indices array.
+                    indices = cgutils.alloca_once(
+                        builder, zero.type,
+                        size=context.get_constant(types.intp, arrty.ndim))
+
+                    for dim in range(arrty.ndim):
+                        idxptr = cgutils.gep(builder, indices, dim)
+                        builder.store(zero, idxptr)
+
+                    self.indices = indices
+
             def iternext_specific(self, context, builder, arrty, arr, result):
+                zero = context.get_constant(types.intp, 0)
+                one = context.get_constant(types.intp, 1)
+
+                ndim = arrty.ndim
                 nitems = arr.nitems
 
                 index = builder.load(self.index)
@@ -608,9 +811,19 @@ def make_array_flat_cls(flatiterty):
                 with cgutils.if_likely(builder, is_valid):
                     ptr = builder.load(self.pointer)
                     value = context.unpack_value(builder, arrty.dtype, ptr)
-                    result.yield_(value)
+                    if kind == 'flat':
+                        result.yield_(value)
+                    else:
+                        # ndenumerate(): fetch and increment indices
+                        indices = self.indices
+                        idxvals = [builder.load(cgutils.gep(builder, indices, dim))
+                                   for dim in range(ndim)]
+                        idxtuple = cgutils.pack_array(builder, idxvals)
+                        result.yield_(
+                            cgutils.make_anonymous_struct(builder, [idxtuple, value]))
+                        _increment_indices_array(context, builder, arrty, arr, indices)
 
-                    index = builder.add(index, context.get_constant(types.intp, 1))
+                    index = builder.add(index, one)
                     builder.store(index, self.index)
                     ptr = cgutils.pointer_add(builder, ptr, self.stride)
                     builder.store(ptr, self.pointer)
@@ -620,19 +833,19 @@ def make_array_flat_cls(flatiterty):
     else:
         class FlatIter(cgutils.Structure):
             """
-            Generic .flat() implementation for non-contiguous arrays.
+            Generic .flat() / .ndenumerate() implementation for
+            non-contiguous arrays.
             It keeps track of pointers along each dimension in order to
             minimize computations.
             """
             _fields = [('array', types.CPointer(array_type)),
                        ('pointers', types.CPointer(types.CPointer(dtype))),
                        ('indices', types.CPointer(types.intp)),
-                       ('empty', types.CPointer(types.boolean)),
+                       ('exhausted', types.CPointer(types.boolean)),
                        ]
 
             def init_specific(self, context, builder, arrty, arr):
                 zero = context.get_constant(types.intp, 0)
-                one = context.get_constant(types.intp, 1)
                 data = arr.data
                 ndim = arrty.ndim
                 shapes = cgutils.unpack_tuple(builder, arr.shape, ndim)
@@ -644,33 +857,25 @@ def make_array_flat_cls(flatiterty):
                                                size=context.get_constant(types.intp,
                                                                          arrty.ndim))
                 strides = cgutils.unpack_tuple(builder, arr.strides, ndim)
-                empty = cgutils.alloca_once_value(builder, cgutils.false_byte)
+                exhausted = cgutils.alloca_once_value(builder, cgutils.false_byte)
 
-                # Initialize each dimension with the next index and pointer
-                # values.  For the last (inner) dimension, this is 0 and the
-                # start pointer, for the other dimensions, this is 1 and the
-                # pointer to the next subarray after start.
+                # Initialize indices and pointers with their start values.
                 for dim in range(ndim):
                     idxptr = cgutils.gep(builder, indices, dim)
                     ptrptr = cgutils.gep(builder, pointers, dim)
-                    if dim == ndim - 1:
-                        builder.store(zero, idxptr)
-                        builder.store(data, ptrptr)
-                    else:
-                        p = cgutils.pointer_add(builder, data, strides[dim])
-                        builder.store(p, ptrptr)
-                        builder.store(one, idxptr)
+                    builder.store(data, ptrptr)
+                    builder.store(zero, idxptr)
                     # 0-sized dimensions really indicate an empty array,
                     # but we have to catch that condition early to avoid
                     # a bug inside the iteration logic (see issue #846).
                     dim_size = shapes[dim]
                     dim_is_empty = builder.icmp(lc.ICMP_EQ, dim_size, zero)
                     with cgutils.if_unlikely(builder, dim_is_empty):
-                        builder.store(cgutils.true_byte, empty)
+                        builder.store(cgutils.true_byte, exhausted)
 
                 self.indices = indices
                 self.pointers = pointers
-                self.empty = empty
+                self.exhausted = exhausted
 
             def iternext_specific(self, context, builder, arrty, arr, result):
                 ndim = arrty.ndim
@@ -682,59 +887,57 @@ def make_array_flat_cls(flatiterty):
 
                 zero = context.get_constant(types.intp, 0)
                 one = context.get_constant(types.intp, 1)
-                minus_one = context.get_constant(types.intp, -1)
-                result.set_valid(True)
 
-                bbcont = cgutils.append_basic_block(builder, 'continued')
                 bbend = cgutils.append_basic_block(builder, 'end')
 
                 # Catch already computed iterator exhaustion
-                is_empty = cgutils.as_bool_bit(builder, builder.load(self.empty))
-                with cgutils.if_unlikely(builder, is_empty):
+                is_exhausted = cgutils.as_bool_bit(
+                    builder, builder.load(self.exhausted))
+                with cgutils.if_unlikely(builder, is_exhausted):
                     result.set_valid(False)
                     builder.branch(bbend)
+                result.set_valid(True)
 
                 # Current pointer inside last dimension
-                last_ptr = cgutils.alloca_once(builder, data.type)
+                last_ptr = cgutils.gep(builder, pointers, ndim - 1)
+                ptr = builder.load(last_ptr)
+                value = context.unpack_value(builder, arrty.dtype, ptr)
+                if kind == 'flat':
+                    result.yield_(value)
+                else:
+                    # ndenumerate() => yield (indices, value)
+                    idxvals = [builder.load(cgutils.gep(builder, indices, dim))
+                               for dim in range(ndim)]
+                    idxtuple = cgutils.pack_array(builder, idxvals)
+                    result.yield_(
+                        cgutils.make_anonymous_struct(builder, [idxtuple, value]))
 
-                # Walk from inner dimension to outer
+                # Update indices and pointers by walking from inner
+                # dimension to outer.
                 for dim in reversed(range(ndim)):
                     idxptr = cgutils.gep(builder, indices, dim)
-                    idx = builder.load(idxptr)
+                    idx = builder.add(builder.load(idxptr), one)
 
                     count = shapes[dim]
                     stride = strides[dim]
                     in_bounds = builder.icmp(lc.ICMP_SLT, idx, count)
                     with cgutils.if_likely(builder, in_bounds):
-                        # Index is valid => we point to the right slot
+                        # Index is valid => pointer can simply be incremented.
+                        builder.store(idx, idxptr)
                         ptrptr = cgutils.gep(builder, pointers, dim)
                         ptr = builder.load(ptrptr)
-                        builder.store(ptr, last_ptr)
-                        # Compute next index and pointer for this dimension
-                        next_ptr = cgutils.pointer_add(builder, ptr, stride)
-                        builder.store(next_ptr, ptrptr)
-                        next_idx = builder.add(idx, one)
-                        builder.store(next_idx, idxptr)
-                        # Reset inner dimensions
+                        ptr = cgutils.pointer_add(builder, ptr, stride)
+                        builder.store(ptr, ptrptr)
+                        # Reset pointers in inner dimensions
                         for inner_dim in range(dim + 1, ndim):
-                            idxptr = cgutils.gep(builder, indices, inner_dim)
                             ptrptr = cgutils.gep(builder, pointers, inner_dim)
-                            # Compute next index and pointer for this dimension
-                            inner_ptr = cgutils.pointer_add(builder, ptr,
-                                                            strides[inner_dim])
-                            builder.store(inner_ptr, ptrptr)
-                            builder.store(one, idxptr)
-                        builder.branch(bbcont)
+                            builder.store(ptr, ptrptr)
+                        builder.branch(bbend)
+                    # Reset index and continue with next dimension
+                    builder.store(zero, idxptr)
 
-                # End of array => skip to end
-                result.set_valid(False)
-                builder.branch(bbend)
-
-                builder.position_at_end(bbcont)
-                # After processing of indices and pointers: fetch value.
-                ptr = builder.load(last_ptr)
-                value = context.unpack_value(builder, arrty.dtype, ptr)
-                result.yield_(value)
+                # End of array
+                builder.store(cgutils.true_byte, self.exhausted)
                 builder.branch(bbend)
 
                 builder.position_at_end(bbend)
@@ -774,3 +977,83 @@ def iternext_numpy_flatiter(context, builder, sig, args, result):
     arr = arrcls(context, builder, value=builder.load(flatiter.array))
 
     flatiter.iternext_specific(context, builder, arrty, arr, result)
+
+
+@builtin
+@implement(numpy.ndenumerate, types.Kind(types.Array))
+def make_array_ndenumerate(context, builder, sig, args):
+    arrty, = sig.args
+    arr, = args
+    nditercls = make_array_ndenumerate_cls(types.NumpyNdEnumerateType(arrty))
+    nditer = nditercls(context, builder)
+
+    arrayptr = cgutils.alloca_once_value(builder, arr)
+    nditer.array = arrayptr
+
+    arrcls = context.make_array(arrty)
+    arr = arrcls(context, builder, ref=arrayptr)
+
+    nditer.init_specific(context, builder, arrty, arr)
+
+    return nditer._getvalue()
+
+
+@builtin
+@implement('iternext', types.Kind(types.NumpyNdEnumerateType))
+@iternext_impl
+def iternext_numpy_nditer(context, builder, sig, args, result):
+    [nditerty] = sig.args
+    [nditer] = args
+
+    nditercls = make_array_ndenumerate_cls(nditerty)
+    nditer = nditercls(context, builder, value=nditer)
+
+    arrty = nditerty.array_type
+    arrcls = context.make_array(arrty)
+    arr = arrcls(context, builder, value=builder.load(nditer.array))
+
+    nditer.iternext_specific(context, builder, arrty, arr, result)
+
+
+@builtin
+@implement(numpy.ndindex, types.VarArg(types.Kind(types.Integer)))
+def make_array_ndindex(context, builder, sig, args):
+    """ndindex(*shape)"""
+    shape = [context.cast(builder, arg, argty, types.intp)
+             for argty, arg in zip(sig.args, args)]
+
+    nditercls = make_ndindex_cls(types.NumpyNdIndexType(len(shape)))
+    nditer = nditercls(context, builder)
+    nditer.init_specific(context, builder, shape)
+
+    return nditer._getvalue()
+
+@builtin
+@implement(numpy.ndindex, types.Kind(types.UniTuple))
+def make_array_ndindex(context, builder, sig, args):
+    """ndindex(shape)"""
+    ndim = sig.return_type.ndim
+    idxty = sig.args[0].dtype
+    tup = args[0]
+
+    shape = cgutils.unpack_tuple(builder, tup, ndim)
+    shape = [context.cast(builder, idx, idxty, types.intp)
+             for idx in shape]
+
+    nditercls = make_ndindex_cls(types.NumpyNdIndexType(len(shape)))
+    nditer = nditercls(context, builder)
+    nditer.init_specific(context, builder, shape)
+
+    return nditer._getvalue()
+
+@builtin
+@implement('iternext', types.Kind(types.NumpyNdIndexType))
+@iternext_impl
+def iternext_numpy_ndindex(context, builder, sig, args, result):
+    [nditerty] = sig.args
+    [nditer] = args
+
+    nditercls = make_ndindex_cls(nditerty)
+    nditer = nditercls(context, builder, value=nditer)
+
+    nditer.iternext_specific(context, builder, result)
