@@ -67,27 +67,27 @@ class Overloads(object):
                 return ver
 
             # As generic type
-            nargs_matches = len(ver.signature.args) == len(sig.args)
-            varargs_matches = (ver.signature.args and
-                               ver.signature.args[-1] == types.VarArg)
-            if nargs_matches or varargs_matches:
-                match = True
-                for formal, actual in zip(ver.signature.args, sig.args):
-                    if formal == types.VarArg:
-                        # vararg argument matches everything
-                        break
-
-                    match = self._match(formal, actual)
-                    if not match:
-                        break
-
-                if match:
-                    return ver
+            if self._match_arglist(ver.signature.args, sig.args):
+                return ver
 
         raise NotImplementedError(self, sig)
 
-    @staticmethod
-    def _match(formal, actual):
+    def _match_arglist(self, formal_args, actual_args):
+        if formal_args and isinstance(formal_args[-1], types.VarArg):
+            formal_args = (
+                formal_args[:-1] +
+                (formal_args[-1].dtype,) * (len(actual_args) - len(formal_args) + 1))
+
+        if len(formal_args) != len(actual_args):
+            return False
+
+        for formal, actual in zip(formal_args, actual_args):
+            if not self._match(formal, actual):
+                return False
+
+        return True
+
+    def _match(self, formal, actual):
         if formal == actual:
             # formal argument matches actual arguments
             return True
@@ -169,6 +169,13 @@ class BaseContext(object):
         name = "CallTemplate(%s)" % fndesc.mangled_name
         self.users[func] = type(name, baseclses, glbls)
 
+    def add_user_function(self, func, fndesc, libs=()):
+        if func not in self.users:
+            msg = "{func} is not a registered user function"
+            raise KeyError(msg.format(func=func))
+        imp = user_function(func, fndesc, libs)
+        self.defns[func].append(imp)
+
     def insert_class(self, cls, attrs):
         clsty = types.Object(cls)
         for name, vtype in utils.iteritems(attrs):
@@ -224,6 +231,8 @@ class BaseContext(object):
         for ak, av in zip(fndesc.args, self.get_arguments(fn)):
             av.name = "arg.%s" % ak
         fn.args[0].name = ".ret"
+        if fndesc.inline:
+            fn.attributes.add('alwaysinline')
         return fn
 
     def declare_external_function(self, module, fndesc):
@@ -422,8 +431,13 @@ class BaseContext(object):
                       for i, v in enumerate(val)]
             return Constant.struct(consts)
 
+        elif isinstance(ty, types.Record):
+            consts = [self.get_constant(types.int8, b)
+                      for b in bytearray(val.tostring())]
+            return Constant.array(consts[0].type, consts)
+
         else:
-            raise NotImplementedError(ty)
+            raise NotImplementedError("%s as constant unsupported" % ty)
 
     def get_constant(self, ty, val):
         assert not self.is_struct_type(ty)
@@ -564,6 +578,13 @@ class BaseContext(object):
             return builder.load(val)
         return val
 
+    def get_returned_value(self, builder, ty, val):
+        """
+        Return value representation to local value representation
+        """
+        # Same as get_argument_value
+        return self.get_argument_value(builder, ty, val)
+
     def get_return_value(self, builder, ty, val):
         """
         Local value representation to return type representation
@@ -765,11 +786,13 @@ class BaseContext(object):
                 tup = builder.insert_value(tup, val, idx)
             return tup
 
-        elif (types.is_int_tuple(toty) and types.is_int_tuple(fromty) and
-                len(toty) == len(fromty)):
+        elif (isinstance(fromty, (types.UniTuple, types.Tuple)) and
+                  isinstance(toty, (types.UniTuple, types.Tuple)) and
+                      len(toty) == len(fromty)):
+
             olditems = cgutils.unpack_tuple(builder, val, len(fromty))
-            items = [self.cast(builder, i, t, toty.dtype)
-                     for i, t in zip(olditems, fromty.types)]
+            items = [self.cast(builder, i, f, t)
+                     for i, f, t in zip(olditems, fromty, toty)]
             tup = self.get_constant_undef(toty)
             for idx, val in enumerate(items):
                 tup = builder.insert_value(tup, val, idx)
@@ -828,12 +851,11 @@ class BaseContext(object):
         if typ in types.integer_domain:
             return builder.icmp(lc.ICMP_NE, val, Constant.null(val.type))
         elif typ in types.real_domain:
-            return builder.fcmp(lc.FCMP_ONE, val, Constant.real(val.type, 0))
+            return builder.fcmp(lc.FCMP_UNE, val, Constant.real(val.type, 0))
         elif typ in types.complex_domain:
             cmplx = self.make_complex(typ)(self, builder, val)
-            fty = types.float32 if typ == types.complex64 else types.float64
-            real_istrue = self.is_true(builder, fty, cmplx.real)
-            imag_istrue = self.is_true(builder, fty, cmplx.imag)
+            real_istrue = self.is_true(builder, typ.underlying_float, cmplx.real)
+            imag_istrue = self.is_true(builder, typ.underlying_float, cmplx.imag)
             return builder.or_(real_istrue, imag_istrue)
         raise NotImplementedError("is_true", val, typ)
 
@@ -844,15 +866,17 @@ class BaseContext(object):
         """
         assert env is None
         retty = callee.args[0].type.pointee
-        retval = cgutils.alloca_once(builder, retty)
+        retvaltmp = cgutils.alloca_once(builder, retty)
         # initialize return value
-        builder.store(lc.Constant.null(retty), retval)
+        builder.store(lc.Constant.null(retty), retvaltmp)
         args = [self.get_value_as_argument(builder, ty, arg)
                 for ty, arg in zip(argtys, args)]
-        realargs = [retval] + list(args)
+        realargs = [retvaltmp] + list(args)
         code = builder.call(callee, realargs)
         status = self.get_return_status(builder, code)
-        return status, builder.load(retval)
+        retval = builder.load(retvaltmp)
+        out = self.get_returned_value(builder, resty, retval)
+        return status, out
 
     def call_external_function(self, builder, callee, argtys, args):
         args = [self.get_value_as_argument(builder, ty, arg)
@@ -870,13 +894,13 @@ class BaseContext(object):
         status = Status(code=code, ok=ok, err=err, exc=exc, none=none)
         return status
 
-    def call_function_pointer(self, builder, funcptr, signature, args):
+    def call_function_pointer(self, builder, funcptr, signature, args, cconv=None):
         retty = self.get_value_type(signature.return_type)
         fnty = Type.function(retty, [a.type for a in args])
         fnptrty = Type.pointer(fnty)
         addr = self.get_constant(types.intp, funcptr)
         ptr = builder.inttoptr(addr, fnptrty)
-        return builder.call(ptr, args)
+        return builder.call(ptr, args, cconv=cconv)
 
     def call_class_method(self, builder, func, signature, args):
         api = self.get_python_api(builder)
@@ -959,14 +983,9 @@ class BaseContext(object):
                                              sig.return_type, flags,
                                              locals=locals)
 
-            # Set to linkonce one-definition-rule to prevent multiple
-            # definitions from raising errors (this can happen with
-            # nested compile_internal()).
+            # Allow inlining the function inside callers.
             codegen.add_linking_library(cres.library)
-            llvm_func = cres.library.get_function(cres.fndesc.llvm_func_name)
-            llvm_func.linkage = 'linkonce_odr'
             fndesc = cres.fndesc
-
             self.cached_internal_func[cache_key] = fndesc
 
         # Add call to the generated function
@@ -1016,12 +1035,11 @@ class BaseContext(object):
 
         # Handle data
         if self.is_struct_type(typ.dtype):
-            # FIXME
-            raise TypeError("Do not support structure dtype as constant "
-                            "array, yet.")
-
-        values = [self.get_constant(typ.dtype, flat[i])
-                  for i in range(flat.size)]
+            values = [self.get_constant_struct(builder, typ.dtype, flat[i])
+                      for i in range(flat.size)]
+        else:
+            values = [self.get_constant(typ.dtype, flat[i])
+                      for i in range(flat.size)]
 
         lldtype = values[0].type
         consts = Constant.array(lldtype, values)
