@@ -1213,28 +1213,14 @@ def caster(restype):
 
     return _cast
 
-
-builtin(caster(types.int8))
-builtin(caster(types.int16))
-builtin(caster(types.int32))
-builtin(caster(types.int64))
-
-builtin(caster(types.uint8))
-builtin(caster(types.uint16))
-builtin(caster(types.uint32))
-builtin(caster(types.uint64))
-
-builtin(caster(types.float32))
-builtin(caster(types.float64))
-
-builtin(caster(types.complex64))
-builtin(caster(types.complex128))
+for tp in types.number_domain:
+    builtin(caster(tp))
 
 
 #-------------------------------------------------------------------------------
 
 @builtin
-@implement(max, types.VarArg)
+@implement(max, types.VarArg(types.Any))
 def max_impl(context, builder, sig, args):
     argtys = sig.args
     for a in argtys:
@@ -1259,7 +1245,7 @@ def max_impl(context, builder, sig, args):
 
 
 @builtin
-@implement(min, types.VarArg)
+@implement(min, types.VarArg(types.Any))
 def min_impl(context, builder, sig, args):
     argtys = sig.args
     for a in argtys:
@@ -1327,7 +1313,7 @@ def float_impl(context, builder, sig, args):
 
 
 @builtin
-@implement(complex, types.VarArg)
+@implement(complex, types.VarArg(types.Any))
 def complex_impl(context, builder, sig, args):
     complex_type = sig.return_type
     float_type = complex_type.underlying_float
