@@ -1,11 +1,13 @@
 from __future__ import print_function
-import numba.unittest_support as unittest
-from numba.compiler import compile_isolated, Flags
-from numba import typeinfer, lowering, types, utils
-from .support import TestCase
+
 import itertools
 import functools
 
+import numba.unittest_support as unittest
+from numba.compiler import compile_isolated, Flags
+from numba import jit, typeof, lowering, types, utils
+from .support import TestCase
+
 enable_pyobj_flags = Flags()
 enable_pyobj_flags.set("enable_pyobject")
 
@@ -56,6 +58,12 @@ def enumerate_usecase():
         result += i * j
     return result
 
+def enumerate_start_usecase():
+    result = 0
+    for i, j in enumerate((1., 2.5, 3.), 42):
+        result += i * j
+    return result
+
 def filter_usecase(x, filter_func):
     return filter(filter_func, x)
 
@@ -65,6 +73,9 @@ def float_usecase(x):
 def format_usecase(x, y):
     return x.format(y)
 
+def globals_usecase():
+    return globals()
+
 def hex_usecase(x):
     return hex(x)
 
@@ -87,12 +98,18 @@ def max_usecase1(x, y):
 def max_usecase2(x, y):
     return max([x, y])
 
+def max_usecase3(x):
+    return max(x)
+
 def min_usecase1(x, y):
     return min(x, y)
 
 def min_usecase2(x, y):
     return min([x, y])
 
+def min_usecase3(x):
+    return min(x)
+
 def oct_usecase(x):
     return oct(x)
 
@@ -129,6 +146,7 @@ def zip_1_usecase():
         result += i
     return result
 
+
 def zip_3_usecase():
     result = 0
     for i, j, k in zip((1, 2), (3, 4, 5), (6.7, 8.9)):
@@ -136,6 +154,14 @@ def zip_3_usecase():
     return result
 
 
+def pow_op_usecase(x, y):
+    return x ** y
+
+
+def pow_usecase(x, y):
+    return pow(x, y)
+
+
 class TestBuiltins(TestCase):
 
     def run_nullary_func(self, pyfunc, flags):
@@ -157,7 +183,9 @@ class TestBuiltins(TestCase):
         for x in [-1.1, 0.0, 1.1]:
             self.assertPreciseEqual(cfunc(x), pyfunc(x), prec='single')
 
-        complex_values = [-1.1 + 0.5j, 0.0 + 0j, 1.1 + 3j]
+        complex_values = [-1.1 + 0.5j, 0.0 + 0j, 1.1 + 3j,
+                          float('inf') + 1j * float('nan'),
+                          float('nan') - 1j * float('inf')]
         cr = compile_isolated(pyfunc, (types.complex64,), flags=flags)
         cfunc = cr.entry_point
         for x in complex_values:
@@ -205,6 +233,14 @@ class TestBuiltins(TestCase):
         cfunc = cr.entry_point
         for x in [-1, 0, 1]:
             self.assertPreciseEqual(cfunc(x), pyfunc(x))
+        cr = compile_isolated(pyfunc, (types.float64,), flags=flags)
+        cfunc = cr.entry_point
+        for x in [0.0, -0.0, 1.5, float('inf'), float('nan')]:
+            self.assertPreciseEqual(cfunc(x), pyfunc(x))
+        cr = compile_isolated(pyfunc, (types.complex128,), flags=flags)
+        cfunc = cr.entry_point
+        for x in [complex(0, float('inf')), complex(0, float('nan'))]:
+            self.assertPreciseEqual(cfunc(x), pyfunc(x))
 
     def test_bool_npm(self):
         self.test_bool(flags=no_pyobj_flags)
@@ -275,6 +311,12 @@ class TestBuiltins(TestCase):
     def test_enumerate_npm(self):
         self.test_enumerate(flags=no_pyobj_flags)
 
+    def test_enumerate_start(self, flags=enable_pyobj_flags):
+        self.run_nullary_func(enumerate_start_usecase, flags)
+
+    def test_enumerate_start_npm(self):
+        self.test_enumerate_start(flags=no_pyobj_flags)
+
     def test_filter(self, flags=enable_pyobj_flags):
         pyfunc = filter_usecase
         cr = compile_isolated(pyfunc, (types.Dummy('list'),
@@ -340,6 +382,29 @@ class TestBuiltins(TestCase):
         with self.assertTypingError():
             self.test_format(flags=no_pyobj_flags)
 
+    def test_globals(self, flags=enable_pyobj_flags):
+        pyfunc = globals_usecase
+        cr = compile_isolated(pyfunc, (), flags=flags)
+        cfunc = cr.entry_point
+        g = cfunc()
+        self.assertIs(g, globals())
+
+    def test_globals_npm(self):
+        with self.assertTypingError():
+            self.test_globals(flags=no_pyobj_flags)
+
+    def test_globals_jit(self, **jit_flags):
+        # Issue #416: weird behaviour of globals() in combination with
+        # the @jit decorator.
+        pyfunc = globals_usecase
+        jitted = jit(**jit_flags)(pyfunc)
+        self.assertIs(jitted(), globals())
+        self.assertIs(jitted(), globals())
+
+    def test_globals_jit_npm(self):
+        with self.assertTypingError():
+            self.test_globals_jit(nopython=True)
+
     def test_hex(self, flags=enable_pyobj_flags):
         pyfunc = hex_usecase
 
@@ -413,7 +478,6 @@ class TestBuiltins(TestCase):
             self.test_map(flags=no_pyobj_flags)
 
     def test_max_1(self, flags=enable_pyobj_flags):
-
         pyfunc = max_usecase1
         cr = compile_isolated(pyfunc, (types.int32, types.int32), flags=flags)
         cfunc = cr.entry_point
@@ -467,6 +531,59 @@ class TestBuiltins(TestCase):
         with self.assertTypingError():
             self.test_min_2(flags=no_pyobj_flags)
 
+    def check_min_max_invalid_types(self, pyfunc, flags=enable_pyobj_flags):
+        cr = compile_isolated(pyfunc, (types.int32, types.Dummy('list')),
+                              flags=flags)
+        cfunc = cr.entry_point
+        cfunc(1, [1])
+
+    def test_max_1_invalid_types(self):
+        # Heterogenous ordering is valid in Python 2
+        if utils.IS_PY3:
+            with self.assertRaises(TypeError):
+                self.check_min_max_invalid_types(max_usecase1)
+        else:
+            self.check_min_max_invalid_types(max_usecase1)
+
+    def test_max_1_invalid_types_npm(self):
+        with self.assertTypingError():
+            self.check_min_max_invalid_types(max_usecase1, flags=no_pyobj_flags)
+
+    def test_min_1_invalid_types(self):
+        # Heterogenous ordering is valid in Python 2
+        if utils.IS_PY3:
+            with self.assertRaises(TypeError):
+                self.check_min_max_invalid_types(min_usecase1)
+        else:
+            self.check_min_max_invalid_types(min_usecase1)
+
+    def test_min_1_invalid_types_npm(self):
+        with self.assertTypingError():
+            self.check_min_max_invalid_types(min_usecase1, flags=no_pyobj_flags)
+
+    # Test that max(1) and min(1) fail
+
+    def check_min_max_unary_non_iterable(self, pyfunc, flags=enable_pyobj_flags):
+        cr = compile_isolated(pyfunc, (types.int32,), flags=flags)
+        cfunc = cr.entry_point
+        cfunc(1)
+
+    def test_max_unary_non_iterable(self):
+        with self.assertRaises(TypeError):
+            self.check_min_max_unary_non_iterable(max_usecase3)
+
+    def test_max_unary_non_iterable_npm(self):
+        with self.assertTypingError():
+            self.check_min_max_unary_non_iterable(max_usecase3)
+
+    def test_min_unary_non_iterable(self):
+        with self.assertRaises(TypeError):
+            self.check_min_max_unary_non_iterable(min_usecase3)
+
+    def test_min_unary_non_iterable_npm(self):
+        with self.assertTypingError():
+            self.check_min_max_unary_non_iterable(min_usecase3)
+
     def test_oct(self, flags=enable_pyobj_flags):
         pyfunc = oct_usecase
 
@@ -582,6 +699,33 @@ class TestBuiltins(TestCase):
     def test_zip_0_npm(self):
         self.test_zip_0(flags=no_pyobj_flags)
 
+    def test_pow_op_usecase(self):
+        args = [
+            (2, 3),
+            (2.0, 3),
+            (2, 3.0),
+            (2j, 3.0j),
+        ]
+
+        for x, y in args:
+            cres = compile_isolated(pow_op_usecase, (typeof(x), typeof(y)),
+                                    flags=no_pyobj_flags)
+            r = cres.entry_point(x, y)
+            self.assertPreciseEqual(r, pow_op_usecase(x, y))
+
+    def test_pow_usecase(self):
+        args = [
+            (2, 3),
+            (2.0, 3),
+            (2, 3.0),
+            (2j, 3.0j),
+        ]
+
+        for x, y in args:
+            cres = compile_isolated(pow_usecase, (typeof(x), typeof(y)),
+                                    flags=no_pyobj_flags)
+            r = cres.entry_point(x, y)
+            self.assertPreciseEqual(r, pow_usecase(x, y))
 
 if __name__ == '__main__':
     unittest.main()
