@@ -1,8 +1,11 @@
 from __future__ import print_function, division, absolute_import
 
+import atexit
 import collections
 import functools
 import io
+import itertools
+import threading
 import timeit
 import math
 import sys
@@ -13,12 +16,50 @@ except ImportError:
 
 import numpy
 
+from .six import *
 from numba.config import PYVERSION, MACHINE_BITS
 
 
-INT_TYPES = (int,)
-if PYVERSION < (3, 0):
-    INT_TYPES += (long,)
+IS_PY3 = PYVERSION >= (3, 0)
+
+if IS_PY3:
+    INT_TYPES = (int,)
+    longint = int
+    get_ident = threading.get_ident
+else:
+    import thread
+    INT_TYPES = (int, long)
+    longint = long
+    get_ident = thread.get_ident
+
+try:
+    from inspect import signature as pysignature
+except ImportError:
+    try:
+        from funcsigs import signature as pysignature
+    except ImportError:
+        raise ImportError("please install the 'funcsigs' package "
+                          "('pip install funcsigs')")
+
+
+_shutting_down = False
+
+def _at_shutdown():
+    global _shutting_down
+    _shutting_down = True
+
+atexit.register(_at_shutdown)
+
+def shutting_down(globals=globals):
+    """
+    Whether the interpreter is currently shutting down.
+    For use in finalizers, __del__ methods, and similar; it is advised
+    to early bind this function rather than look it up when calling it,
+    since at shutdown module globals may be cleared.
+    """
+    # At shutdown, the attribute may have been cleared or set to None.
+    v = globals().get('_shutting_down')
+    return v is True or v is None
 
 
 class ConfigOptions(object):
@@ -103,21 +144,60 @@ class UniqueDict(dict):
         super(UniqueDict, self).__setitem__(key, value)
 
 
-# def cache(fn):
-#     @functools.wraps(fn)
-#     def cached_func(self, *args, **kws):
-#         if self in cached_func.cache:
-#             return cached_func.cache[self]
-#         ret = fn(self, *args, **kws)
-#         cached_func.cache[self] = ret
-#         return ret
-#     cached_func.cache = {}
-#     def invalidate(self):
-#         if self in cached_func.cache:
-#             del cached_func.cache[self]
-#     cached_func.invalidate = invalidate
-#
-#     return cached_func
+class NonReentrantLock(object):
+    """
+    A lock class which explicitly forbids reentrancy.
+    """
+
+    def __init__(self):
+        self._lock = threading.Lock()
+        self._owner = None
+
+    def acquire(self):
+        me = get_ident()
+        if me == self._owner:
+            raise RuntimeError("cannot re-acquire lock from same thread")
+        self._lock.acquire()
+        self._owner = me
+
+    def release(self):
+        if self._owner != get_ident():
+            raise RuntimeError("cannot release un-acquired lock")
+        self._owner = None
+        self._lock.release()
+
+    def is_owned(self):
+        """
+        Whether the lock is owned by the current thread.
+        """
+        return self._owner == get_ident()
+
+    __enter__ = acquire
+
+    def __exit__(self, t, v, tb):
+        self.release()
+
+
+# Django's cached_property
+# see https://docs.djangoproject.com/en/dev/ref/utils/#django.utils.functional.cached_property
+
+class cached_property(object):
+    """
+    Decorator that converts a method with a single self argument into a
+    property cached on the instance.
+
+    Optional ``name`` argument allows you to make cached properties of other
+    methods. (e.g.  url = cached_property(get_absolute_url, name='url') )
+    """
+    def __init__(self, func, name=None):
+        self.func = func
+        self.name = name or func.__name__
+
+    def __get__(self, instance, type=None):
+        if instance is None:
+            return self
+        res = instance.__dict__[self.name] = self.func(instance)
+        return res
 
 
 def runonce(fn):
@@ -191,63 +271,7 @@ if PYVERSION < (3, 0):
     RANGE_ITER_OBJECTS += (builtins.xrange,)
 
 
-# Other common python2/3 adaptors
-# Copied from Blaze which borrowed from six
-
-IS_PY3 = PYVERSION >= (3, 0)
-
-if IS_PY3:
-    def dict_iteritems(d):
-        return d.items().__iter__()
-
-    def dict_itervalues(d):
-        return d.values().__iter__()
-
-    def dict_values(d):
-        return list(d.values())
-
-    def dict_keys(d):
-        return list(d.keys())
-
-    def iter_next(it):
-        return it.__next__()
-
-    def func_globals(f):
-        return f.__globals__
-
-    longint = int
-
-    unicode = str
-
-    StringIO = io.StringIO
-
-else:
-    from cStringIO import StringIO
-
-    def dict_iteritems(d):
-        return d.iteritems()
-
-    def dict_itervalues(d):
-        return d.itervalues()
-
-    def dict_values(d):
-        return d.values()
-
-    def dict_keys(d):
-        return d.keys()
-
-    def iter_next(it):
-        return it.next()
-
-    def func_globals(f):
-        return f.func_globals
-
-    longint = long
-
-    unicode = unicode
-
-
-# Backported from Python 3.4
+# Backported from Python 3.4: functools.total_ordering()
 
 def _not_op(op, other):
     # "not a < b" handles "a >= b"
@@ -355,3 +379,144 @@ def total_ordering(cls):
             opfunc.__doc__ = getattr(int, opname).__doc__
             setattr(cls, opname, opfunc)
     return cls
+
+
+# Backported from Python 3.4: weakref.finalize()
+
+from weakref import ref
+
+class finalize:
+    """Class for finalization of weakrefable objects
+
+    finalize(obj, func, *args, **kwargs) returns a callable finalizer
+    object which will be called when obj is garbage collected. The
+    first time the finalizer is called it evaluates func(*arg, **kwargs)
+    and returns the result. After this the finalizer is dead, and
+    calling it just returns None.
+
+    When the program exits any remaining finalizers for which the
+    atexit attribute is true will be run in reverse order of creation.
+    By default atexit is true.
+    """
+
+    # Finalizer objects don't have any state of their own.  They are
+    # just used as keys to lookup _Info objects in the registry.  This
+    # ensures that they cannot be part of a ref-cycle.
+
+    __slots__ = ()
+    _registry = {}
+    _shutdown = False
+    _index_iter = itertools.count()
+    _dirty = False
+    _registered_with_atexit = False
+
+    class _Info:
+        __slots__ = ("weakref", "func", "args", "kwargs", "atexit", "index")
+
+    def __init__(self, obj, func, *args, **kwargs):
+        if not self._registered_with_atexit:
+            # We may register the exit function more than once because
+            # of a thread race, but that is harmless
+            import atexit
+            atexit.register(self._exitfunc)
+            finalize._registered_with_atexit = True
+        info = self._Info()
+        info.weakref = ref(obj, self)
+        info.func = func
+        info.args = args
+        info.kwargs = kwargs or None
+        info.atexit = True
+        info.index = next(self._index_iter)
+        self._registry[self] = info
+        finalize._dirty = True
+
+    def __call__(self, _=None):
+        """If alive then mark as dead and return func(*args, **kwargs);
+        otherwise return None"""
+        info = self._registry.pop(self, None)
+        if info and not self._shutdown:
+            return info.func(*info.args, **(info.kwargs or {}))
+
+    def detach(self):
+        """If alive then mark as dead and return (obj, func, args, kwargs);
+        otherwise return None"""
+        info = self._registry.get(self)
+        obj = info and info.weakref()
+        if obj is not None and self._registry.pop(self, None):
+            return (obj, info.func, info.args, info.kwargs or {})
+
+    def peek(self):
+        """If alive then return (obj, func, args, kwargs);
+        otherwise return None"""
+        info = self._registry.get(self)
+        obj = info and info.weakref()
+        if obj is not None:
+            return (obj, info.func, info.args, info.kwargs or {})
+
+    @property
+    def alive(self):
+        """Whether finalizer is alive"""
+        return self in self._registry
+
+    @property
+    def atexit(self):
+        """Whether finalizer should be called at exit"""
+        info = self._registry.get(self)
+        return bool(info) and info.atexit
+
+    @atexit.setter
+    def atexit(self, value):
+        info = self._registry.get(self)
+        if info:
+            info.atexit = bool(value)
+
+    def __repr__(self):
+        info = self._registry.get(self)
+        obj = info and info.weakref()
+        if obj is None:
+            return '<%s object at %#x; dead>' % (type(self).__name__, id(self))
+        else:
+            return '<%s object at %#x; for %r at %#x>' % \
+                (type(self).__name__, id(self), type(obj).__name__, id(obj))
+
+    @classmethod
+    def _select_for_exit(cls):
+        # Return live finalizers marked for exit, oldest first
+        L = [(f,i) for (f,i) in cls._registry.items() if i.atexit]
+        L.sort(key=lambda item:item[1].index)
+        return [f for (f,i) in L]
+
+    @classmethod
+    def _exitfunc(cls):
+        # At shutdown invoke finalizers for which atexit is true.
+        # This is called once all other non-daemonic threads have been
+        # joined.
+        reenable_gc = False
+        try:
+            if cls._registry:
+                import gc
+                if gc.isenabled():
+                    reenable_gc = True
+                    gc.disable()
+                pending = None
+                while True:
+                    if pending is None or finalize._dirty:
+                        pending = cls._select_for_exit()
+                        finalize._dirty = False
+                    if not pending:
+                        break
+                    f = pending.pop()
+                    try:
+                        # gc is disabled, so (assuming no daemonic
+                        # threads) the following is the only line in
+                        # this function which might trigger creation
+                        # of a new finalizer
+                        f()
+                    except Exception:
+                        sys.excepthook(*sys.exc_info())
+                    assert f not in cls._registry
+        finally:
+            # prevent any more finalizers from executing during shutdown
+            finalize._shutdown = True
+            if reenable_gc:
+                gc.enable()
